import { BehaviorSubject, Observable } from 'rxjs';
import { distinctUntilChanged, map } from 'rxjs/operators';

export class BaseStateService<T> {
  private state$: BehaviorSubject<T>;

  protected get state(): T {
    return this.state$.getValue();
  }

  constructor(initialState: T, private recreateStateOnInitialize: boolean = false) {
    this.initialize(initialState);
  }

  public initialize(initialState: T){
      var state$ = (this as any).state$;

      // di base, si evita di reistanziare il behavior subject, perchè a fronte di una reinizializzazione
      // del servizio senza cambio di route, i componenti che si sono iscritti tramite select
      // devono continuare a ricevere gli updates. Si consente comunque il vecchio comportamento per eventuali
      // ragioni di retrocompatibilità tramite parametro addizionale "recreateStateOnInitialize"
      if (!state$ || this.recreateStateOnInitialize){
          (this as any).state$ = new BehaviorSubject<T>({ ...initialState});
      } else {
        this.setState({ ...initialState });
      }      
  }

  public select<K>(mapFn: (state: T) => K, noClone: boolean = false): Observable<K> {
    return this.state$.asObservable().pipe(
      map((state: T) => mapFn(state)),
      distinctUntilChanged(),
      map(selected => noClone ? selected : this.clone(selected))
    );
  }

  protected setState(newState: Partial<T>) {
    this.state$.next({
      ...this.state,
      ...newState,
    });
  }

  clone<T>(obj: T): T {
    let copy;

    // Handle the 3 simple types, and null or undefined
    if (null == obj || 'object' !== typeof obj) {
        return obj;
    }

    // Handle Date
    if (obj instanceof Date) {
        copy = new Date();
        copy.setTime(obj.getTime());
        return copy;
    }

    // Handle Array
    if (obj instanceof Array) {
        copy = [];
        for (let i = 0, len = obj.length; i < len; i++) {
            copy[i] = this.clone(obj[i]);
        }
        return copy;
    }

    // Handle Object
    if (obj instanceof Object) {
        copy = {};
        for (const attr in obj) {
            if (obj.hasOwnProperty(attr)){
                copy[attr] = this.clone(obj[attr]);
            }
        }
        return copy;
    }

    throw new Error('Unable to copy obj! Its type isn\'t supported.');
  }
}