
import { Injectable } from "@angular/core";
import { EMPTY, map, of, take, tap } from "rxjs";
import { BaseStateService } from "./base-state.service";
import { SchedulazioneEventModel, SchedulazioneEventStatus, SchedulazioneEventType } from "./schedulazione-data";

export enum EventColorizationMode {
    Normal = 0,
    ByDueDate = 1
}

export class GanttColorizationState {
    ColorizationMode: EventColorizationMode; 
    HighlightData: EventHighlightData;
}

export class EventHighlightData {
    HighlightedEventId: number | string;
    FilteredAndSortedResourceIds: number[];
}

export const initialState = {
    ColorizationMode: EventColorizationMode.Normal,
    HighlightData: null
} as GanttColorizationState;

@Injectable()
export class GanttColorizationService extends BaseStateService<GanttColorizationState> {

    private _withinDueDateColor = '#00b300';
    private _missedDueDateColor = '#ff3333';

    private _highlightedEventColor = '#e5f73b';
    private _invalidEventColor = '#ff0000';
    private _lockedEventColor = '#777777';

    constructor() {
        super(initialState);
    }

    initState(){
        console.log('>initState called on GanttColorizationService');
        
        return of(this.state)
            .pipe(
                tap(_ => this.setState({ 
                    ColorizationMode: EventColorizationMode.Normal, 
                    HighlightData: null
                })),
                map(_ => null),
                take(1)
            )
    }

    setColorizationMode(colorizationMode: EventColorizationMode){
        return of(this.state)
            .pipe(
                tap(_ => this.setState({ ColorizationMode: colorizationMode })),
                map(_ => null),
                take(1)
            );
    }

    toggleHighlightEventId(events: SchedulazioneEventModel[], toggleHighlightEventId: number | string = null){
        return of(this.state)
            .pipe(
                map(state => this._getUpdatedHighlightData(events, state.HighlightData, toggleHighlightEventId)),
                tap(updatedHighlightData => this.setState({ HighlightData: updatedHighlightData })),
                map(_ => null),
                take(1)
            );
    }

    highlightBySearch(events: SchedulazioneEventModel[], selectedOdP: string){
        return of(this.state)
            .pipe(
                map(_ => this._getUpdatedHighlightDataBySearch(events, selectedOdP)),
                tap(updatedHighlightData => this.setState({ HighlightData: updatedHighlightData })),
                map(_ => null),
                take(1)
            );
    }

    updateEventsStyle(events: SchedulazioneEventModel[]){
        console.log('***updating events style...');

        return of(this.state)
            .pipe(
                map(state => this._getUpdatedEvents(events, state.ColorizationMode, state.HighlightData)),
                tap(_ => console.log('...done***')),
                map(_ => null),
                take(1)
            );
    }

    private _getUpdatedHighlightDataBySearch(events: SchedulazioneEventModel[], selectedOdP: string){
        if (!selectedOdP){
            return null;
        }

        const odpEvents = events.filter(m => !!m.faseData && m.faseData.OdPCodice === selectedOdP);

        if (odpEvents.length <= 0){
            return null;
        }

        const firstEvent = odpEvents[0];

        const selectedSortedResourcesIds = this.getSortedEventsByNumero(odpEvents, 'asc').map(m => m.faseData.MacchinaId);

        return { HighlightedEventId: firstEvent.id, FilteredAndSortedResourceIds: selectedSortedResourcesIds } as EventHighlightData;
    }
    
    private _getUpdatedHighlightData(events: SchedulazioneEventModel[], currentHighlightData: EventHighlightData, toggleHighlightEventId: number | string){
        // se c'era già un highlight impostato, l'effetto sarà di toglierlo
        if (!!currentHighlightData){
            return null;
        }

        // altrimenti, lo imposto
        const dblClickedEvent = events.find(m => m.id === toggleHighlightEventId);
    
        const odpEvents = events.filter(m => m.faseData.OdPCodice === dblClickedEvent.faseData.OdPCodice);

        const selectedSortedResourcesIds = this.getSortedEventsByNumero(odpEvents, 'asc').map(m => m.faseData.MacchinaId);

        return { HighlightedEventId: toggleHighlightEventId, FilteredAndSortedResourceIds: selectedSortedResourcesIds } as EventHighlightData;
    }
    
/*     private _canBeHighlighted(events: SchedulazioneEventModel[], toggleHighlightEventId: number | string){
        const dblClickedEvent = events.find(m => m.id === toggleHighlightEventId);

        return dblClickedEvent.eventType === SchedulazioneEventType.Normale;
    } */

    private _getUpdatedEvents(events: SchedulazioneEventModel[], colorizationMode: EventColorizationMode, highlightData: EventHighlightData){
        // come prima cosa, se la funzione sta venendo chiamata in risposta ad un'operazione
        // di highlight (impostato o rimosso), aggiorno su ogni evento lo stato di highlight
        let toReturn = this._getEventsWithUpdatedHighlightedFlag(events, highlightData) 

        // poi, ne aggiorno il colore e lo stile, in base al fatto che
        // siano highlighted o meno, ed in base alla colorization mode
        return this._getEventsWithUpdatedStyleAndColor(toReturn, colorizationMode);
    }

    private _getEventsWithUpdatedHighlightedFlag(events: SchedulazioneEventModel[], highlightData: EventHighlightData){
        // innanzitutto, imposto tutti gli eventi come non highlighted
        for(var event of events){
            event.highlighted = false;
        }

        if (!!highlightData){
            // poi, se l'highlight è richiesto
            const eventToHighlight = events.find(m => m.id === highlightData.HighlightedEventId);

            const odpEvents = events.filter(m => m.faseData.OdPCodice === eventToHighlight.faseData.OdPCodice);

            for(var odpEvent of odpEvents){
                // imposto gli eventi dell'odp come highlighted
                odpEvent.highlighted = true;
            }
        }

        return [...events];
    }

    private _getEventsWithUpdatedStyleAndColor(events: SchedulazioneEventModel[], colorizationMode: EventColorizationMode){
        // la colorization non agisce sugli gli eventi di tipo diverso da normale
        const normalEvents = events.filter(m => m.eventType === SchedulazioneEventType.Normale);

        const eventsGroupedByOdp = this.groupBy(normalEvents, m => m.faseData.OdPCodice);

        for(let odp in eventsGroupedByOdp){
            const sortedEventsOdP = this.getSortedEventsByNumero(eventsGroupedByOdp[odp], 'desc');
            const lastEventOdP = sortedEventsOdP[0];

            const missedDueDate = !!lastEventOdP.faseData.OdPDueDate 
                && lastEventOdP.faseData.OdPDueDate < lastEventOdP.endDate;

            for(let odpEvent of sortedEventsOdP){
                odpEvent.eventColor = this._getEventUpdatedColor(colorizationMode, missedDueDate, odpEvent);
                (odpEvent as any).eventStyle = this._getUpdatedEventStyle(odpEvent);
            }
        }

        return [...events];
    }

    private _getEventUpdatedColor(colorizationMode: EventColorizationMode, missedDueDate: boolean, event: SchedulazioneEventModel){
        if (event.highlighted){
            return this._highlightedEventColor;
        } else if (event.eventStatus !== SchedulazioneEventStatus.Valid){
            return this._invalidEventColor;
        } else if (event.isLocked){
            return this._lockedEventColor;
        } else {
            if (colorizationMode === EventColorizationMode.ByDueDate){
                return missedDueDate ? this._missedDueDateColor : this._withinDueDateColor;
            } else {
                return event.originalEventColor;
            }
        } 
    }

    private _getUpdatedEventStyle(event: SchedulazioneEventModel){
        if (event.highlighted){
            return 'hollow';
        } else if (event.eventStatus !== SchedulazioneEventStatus.Valid){
            return 'dashed';
        } else if (event.isLocked){
            return 'line';
        } else {
            return 'border';
        }
    }

    getSortedEventsByNumero(events: SchedulazioneEventModel[], order: string = 'asc'){
        const sortFunc = this.sortFunc<SchedulazioneEventModel>(m => m.faseData.Numero, order);

        // ATTENZIONE: è fondamentale lo spread operator, perchè il metodo sort, 
        // di base restituisce una reference allo stesso array che nel frattempo è stato riordinato.
        // grazie allo spread operator, il metodo restituirà invece una copia dell'array, 
        // con l'ordinamento richiesto applicato ad esso
        return [...events.sort(sortFunc)];
    }

    groupBy<T, K extends keyof any>(arr: T[], key: (i: T) => K){
        return arr.reduce((groups, item) => {
            (groups[key(item)] ||= []).push(item);
            return groups;
          }, {} as Record<K, T[]>);
    }

    sortFunc<TObject>(propFunc: (obj: TObject) => any,  direction: string = 'asc'){
        return (n1: TObject, n2: TObject) => {
            if (propFunc(n1) > propFunc(n2)) {
                return (direction === 'asc') ? 1 : -1;
            }

            if (propFunc(n1) < propFunc(n2)) {
                return (direction === 'asc') ? -1 : 1;
            }

            return 0;
        }
    }
    
}