import { AfterContentInit, AfterViewInit, ChangeDetectorRef, Component, ElementRef, EventEmitter, Input, OnChanges, OnDestroy, OnInit, Output, QueryList, SimpleChanges, ViewChild, ViewChildren, ViewEncapsulation } from '@angular/core';
import { AssignmentModel, CalendarModel, DependencyModel, EventModel, LocaleManager, MenuItem, ProjectModel, ResourceHistogram, ResourceModel, SchedulerPro, TimeSpan } from '@bryntum/schedulerpro';
import { BryntumResourceHistogramComponent, BryntumSchedulerProComponent } from '@bryntum/schedulerpro-angular';
import { schedConfig, histConfig } from './scheduler.config';
import { Subscription, of, filter, tap, switchMap, take, map, from, delay } from 'rxjs';
import { NonWorkingEventsData, SchedulazioneData, SchedulazioneEventModel, SchedulazioneEventType, SchedulazioneResourceModel } from './schedulazione-data';
import { customData } from './data';
import { EventHighlightData, GanttColorizationService } from './gantt-colorization.service';


@Component({
  selector: 'app-gantt',
  templateUrl: './gantt.component.html',
  styleUrls: ['./gantt.component.scss'],
  encapsulation: ViewEncapsulation.None
})
export class GanttComponent implements OnInit, OnDestroy, AfterViewInit {

  private _ganttInitialized = false;

  subscriptions = new Subscription();

  schedulerProConfig = schedConfig;
  histogramConfig = histConfig;

  schedulerPro: SchedulerPro;
  resourceHistogram: ResourceHistogram;

  @ViewChild(BryntumSchedulerProComponent) 
  schedulerProComponent: BryntumSchedulerProComponent;

  @Output()
  deleteSchedulazione = new EventEmitter<number>();

  startDate: Date;
  endDate: Date;

  visibleDate: Date;

  nonWorkingEventsData: NonWorkingEventsData;

  project: ProjectModel = new ProjectModel({
/*     assignmentStore : {
      useRawData : true,
      data       : []
    },
    resourceStore : {
        useRawData : true,
        data       : []
    },
    eventStore : {
        useRawData : true,
        data       : []
    }, */
    events: [],
    resources: [],
    assignments: [],
    calendars: [],   
    dependencies: [],
    skipNonWorkingTimeInDurationWhenSchedulingManually: false
  })

  featuresConfig = {
    eventMenu : {
      items : {
        editEvent: false,
        deleteEvent: false,
        copyEvent: false,
        cutEvent: false,
        splitEvent: false,
        unassignEvent: false,
          deleteSchedulazioneOdLItem : {
              itemType  : 'delete-schedulazione-odl',
              text : 'Elimina schedulazione OdL',
              icon : 'b-fa b-fa-fw b-fa-trash'
          }
      }
    },
    group: {
      disabled: true
    },//'classeMacchina',
    dependencies: {
      allowCreate: false,
      // Round the corner where line segments connect, similar to 'border-radius: 5px'
      radius : 5
    },
    disabled: {
      disabled: true
    },
    nonWorkingTime: {
      disabled: false,
      maxTimeAxisUnit: 'year'
    },
    resourceNonWorkingTime: {
      disabled: false,
      maxTimeAxisUnit: 'week'
    },
    eventDrag: {
      // in questo modo non è possibile spostare gli eventi su altre risorse
      constrainDragToResource : true,
/*       validatorFn({ eventRecords, newResource }){
        console.log(this);
        console.log(eventRecords);
        console.log(newResource);
        return {
          valid: false,
          message: 'nonono'
        };
      } */
    },
  }

  constructor(private cdr: ChangeDetectorRef, private ganttColorizationServ: GanttColorizationService) { }

  ngOnInit(): void {
    console.log('>ngOnInit called on gantt-component');
  }

  onRefresh(){

    let obj: SchedulazioneData = JSON.parse(JSON.stringify(customData));

    this.subscriptions.add(
      of(obj)
        .pipe(
          // mi interessano solo le emissioni diverse da NULL
          filter(data => !!data),
          switchMap(data => from(this._initializeGantt(data))),
          switchMap(_ => this._updateEventsStyleAndResources('test'))
        )
        .subscribe(_ => this.cdr.detectChanges())
    );
  }

  ngAfterViewInit(): void {
    console.log('>ngAfterViewInit called on gantt-component');

    setTimeout(() =>{
      // questo riferimento serve per operare direttamente
      // sull'oggetto scheduler da questo componente 
      // (ad es, per impostare il filtro su risorse..)
      this.schedulerPro = this.schedulerProComponent.instance;

      this.onRefresh()
    });
  }

  private async _initializeGantt(data: SchedulazioneData){
    console.log('***gantt inizialization started...');
    //this.schedulerPro.suspendRefresh();
    this._ganttInitialized = false;

    this.visibleDate = new Date(data.MinDate);

    this.startDate = new Date(data.MinDate);
    this.endDate = new Date(data.MaxDate);

/*     this.project.assignments = [];
    this.project.dependencies = [];
    this.project.events = [];
    this.project.resources = [];
    this.project.calendars = []; */

    this.project.calendars = data.Calendars;
    this.project.resources = data.Resources;
    this.project.events = data.Events;
    this.project.dependencies = data.Dependencies;
    this.project.assignments = data.Assignments;

    this.project.calendar = data.SchedulingWindowCalendarId;

    this.nonWorkingEventsData = data.NonWorkingEventsData;
  
    await this.project.commitAsync();

    this._ganttInitialized = true;

    //this.schedulerPro.resumeRefresh(true);
    console.log('...completed***');
  }

  private _updateEventsStyleAndResources(caller: string){
    // non deve mai essere eseguito se non è ancora stata completata l'inizializzzione..
    if (!this._ganttInitialized){
      return of(null);
    }
    
    const events = this._getEvents();

    console.log('>_updateEventsStyleAndResources related to ' + events.length + ' events was called on gantt-component by ' + caller);

    return this.ganttColorizationServ.updateEventsStyle(events)
      .pipe(
        switchMap(_ => this.ganttColorizationServ.select(m => m.HighlightData)),
        take(1),
        tap(highlightData => this._applyResourcesSortingAndFiltering(events, highlightData)),
        map(_ => null),
        take(1)
      );
  }

  private _applyResourcesSortingAndFiltering(events: SchedulazioneEventModel[], highlightData: EventHighlightData){
    this.schedulerPro.resourceStore.clearFilters();
    
    if (!!highlightData){
      // se i dati di highlight sono presenti, significa che questo
      // è stato appena abilitato: applico il filtro che evidenzierà tutte e sole le macchine relative all'OdP corrente           
      this.schedulerPro.resourceStore.filter(resource => highlightData.FilteredAndSortedResourceIds.includes(resource.id));
      this.schedulerPro.resourceStore.sort((res1, res2) => highlightData.FilteredAndSortedResourceIds.indexOf(res1.id) >= highlightData.FilteredAndSortedResourceIds.indexOf(res2.id) ? 1 : -1);
    } else {
      // altrimenti, se tali dati sono NULL, e il componente è già stato inizializzato
      // significa che l'highlight è appena stato tolto: riporto i filtri nello stato iniziale
      const fasiNormali = (this.project.events as SchedulazioneEventModel[]).filter(m => m.eventType === SchedulazioneEventType.Normale);

      this.schedulerPro.resourceStore.filter(resource => !!fasiNormali.find(m => m.faseData.MacchinaId === resource.id));
      this.schedulerPro.resourceStore.sort((res1, res2) => this._resourcesSortFunc(res1, res2));
    }
  }

  private _resourcesSortFunc(res1: SchedulazioneResourceModel, res2: SchedulazioneResourceModel){
    // di base si ordina in base al numero "ordinamento" che deriva dalla classe,
    // a parità di numero di ordinamento si va in ordine alfabetico
    if (res1.ordinamento > res2.ordinamento){
      return 1;
    } else if (res1.ordinamento < res2.ordinamento){
      return -1;
    } else {
      return (res1.name ?? '').toUpperCase() >= (res2.name ?? '').toUpperCase() ? 1 : -1;
    }
  }

  private _getEvents(){
    return this.project.eventStore.getEvents({
      startDate: this.getMinDate(),
      endDate: this.getMaxDate()
    }) as any as SchedulazioneEventModel[];
  }

  getMinDate(){
      return new Date(1990, 0, 1)
  }

  getMaxDate(){
      return new Date(2300, 0, 1);
  }
  
  ngOnDestroy(){
    this.subscriptions.unsubscribe();
  }

  getEventStyle(){
    return schedConfig.eventStyle as any;
  }

  cancelCallback(name: string, args: any){
    console.log(name + ' called');

    args.returnValue = false;
  }

}
